"""
Command line script to calculate the transformation matrix to fit the implant
"""

import argparse

from aind_mri_targeting.implant_rotations import fit_implant_to_mri_from_files


def parse_args():
    parser = argparse.ArgumentParser(
        description="Calculate the volume transformation matrix of an implant"
    )
    parser.add_argument("segmentation", help="path to the segmentation file")
    parser.add_argument("hole_dir", help="directory containing the hole files")
    parser.add_argument("output", help="path to the output file")
    parser.add_argument(
        "-F",
        "--forward",
        action="store_true",
        default=False,
        help="save the forward transformation matrix instead of the inverse",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="force overwrite",
    )
    parser.add_argument("-m", "--mouse", default=None, help="mouse ID")
    return parser.parse_args()


def main():
    args = parse_args()
    fit_implant_to_mri_from_files(
        args.segmentation,
        args.hole_dir,
        save_name=args.output,
        save_inverse=not args.forward,
        force=args.force,
        mouse_name=args.mouse,
    )
    return 0


if __name__ == "__main__":
    main()



"""Module for implant rotations."""

import re
from pathlib import Path

import SimpleITK as sitk
import trimesh
from aind_mri_utils import coordinate_systems as cs
from aind_mri_utils.implant import fit_implant_to_mri, make_hole_seg_dict
from aind_mri_utils.rotations import rotation_matrix_to_sitk


def find_hole_files(directory_or_files, hole_file_pattern=r"Hole(\d+).obj"):
    """
    Find all hole files in a directory or from a list of files and return a
    dictionary mapping hole numbers to file paths.

    Parameters
    ----------
    directory_or_files : Path, str or iterable
        Either the directory path containing the hole files or an iterable of
        file paths.
    hole_file_pattern : str, optional
        The regular expression pattern to match hole file names. Default is
        r"Hole(\\d+).obj". Must contain a single group to extract the hole
        number.

    Returns
    -------
    dict
        A dictionary where the keys are hole numbers (int) extracted from the
        file names and the values are the corresponding file paths (Path
        objects).

    Raises
    ------
    ValueError
        If `directory_or_files` is a string or Path but is not a valid
        directory.

    Examples
    --------
    >>> find_hole_files("/path/to/directory")
    {1: PosixPath('/path/to/directory/Hole1.obj'),
    2: PosixPath('/path/to/directory/Hole2.obj')}

    >>> find_hole_files(["/path/to/directory/Hole1.obj",
        "/path/to/directory/Hole2.obj"])
    {1: PosixPath('/path/to/directory/Hole1.obj'),
    2: PosixPath('/path/to/directory/Hole2.obj')}
    """
    # Is it a directory or a list of files?
    if isinstance(directory_or_files, str):
        directory_or_files = Path(directory_or_files)
    if isinstance(directory_or_files, Path):
        if not directory_or_files.is_dir():
            raise ValueError(f"{directory_or_files} is not a directory.")
        files = directory_or_files.iterdir()
    else:
        files = directory_or_files

    hole_files = {}
    for file in files:
        if not isinstance(file, Path):
            file = Path(file)
        match = re.match(hole_file_pattern, file.name)
        if match:
            hole_files[int(match.group(1))] = file

    return hole_files


def load_convert_trimesh(file, source_cs="ASR", target_cs="LPS"):
    """
    Load a trimesh object from a file and convert its vertices from one
    coordinate system to another.

    Parameters
    ----------
    file : Path or str
        The file path to the trimesh object.
    source_cs : str, optional
        The coordinate system of the vertices in the file. Default is "ASR".
    target_cs : str, optional
        The coordinate system to convert the vertices to. Default is "LPS".

    Returns
    -------
    trimesh.Trimesh
        The trimesh object with vertices converted to the target coordinate
        system.
    """
    mesh = trimesh.load(file)
    if source_cs != target_cs:
        mesh.vertices = cs.convert_coordinate_system(
            mesh.vertices, source_cs, target_cs
        )
    return mesh


def find_load_meshes(
    hole_directory,
    lower_face_file="LowerFace.obj",
    hole_file_kws={},
    load_convert_kws={},
):
    """
    Find all hole files in a directory, load them as trimesh objects and
    convert their vertices to a target coordinate system.

    Parameters
    ----------
    hole_directory : Path or str
        The directory path containing the hole files.
    lower_face_file : str, optional
        The file name of the lower face mesh. Default is "LowerFace.obj".
    hole_file_kws : dict, optional
        Additional keyword arguments to pass to `find_hole_files`. Default is
        an empty dictionary.
    load_convert_kws : dict, optional
        Additional keyword arguments to pass to `load_convert_trimesh`. Default
        is an empty dictionary.

    Returns
    -------
    dict
        A dictionary where the keys are hole numbers (int) extracted from the
        file names and the values are the corresponding trimesh objects with
        vertices converted to the target coordinate system.

    Examples
    --------
    >>> find_load_meshes("/path/to/directory")
    {1: <trimesh.Trimesh>, 2: <trimesh.Trimesh>}

    >>> find_load_meshes("/path/to/dir", source_cs="LPS", target_cs="ASR")
    {1: <trimesh.Trimesh>, 2: <trimesh.Trimesh>}
    """
    hole_files = find_hole_files(hole_directory, **hole_file_kws)
    hole_meshes = {}
    for hole_num, file in hole_files.items():
        hole_meshes[hole_num] = load_convert_trimesh(file, **load_convert_kws)
    hole_meshes[-1] = load_convert_trimesh(
        Path(hole_directory) / lower_face_file, **load_convert_kws
    )
    return hole_meshes


def fit_implant_to_mri_from_files(
    implant_annotations_file,
    hole_directory,
    save_name=None,
    save_inverse=True,
    force=False,
    mouse_name=None,
    find_load_mesh_kws={},
    fit_kws={},
):
    """
    Fit an implant to an MRI volume using hole files and implant annotations.

    Parameters
    ----------
    implant_annotations_file : Path or str
        The file path to the implant annotations.
    hole_directory : Path or str
        The directory path containing the hole files.
    save_name : Path or str, optional
        The file path to save the transformation matrix. Default is None.
    save_inverse : bool, optional
        Whether to save the inverse of the transformation matrix. Default is
        True.
    force : bool, optional
        Whether to overwrite the save file if it already exists. Default is
        False.
    mouse_name : int or str, optional
        The mouse ID. Default is None.
    find_load_mesh_kws : dict, optional
        Additional keyword arguments to pass to `find_load_meshes`. Default is
        an empty dictionary.
    fit_kws : dict, optional
        Additional keyword arguments to pass to `fit_implant_to_mri`. Default
        is an empty dictionary.

    Returns
    -------
    dict
        A dictionary containing the fitted implant and hole rotation matrices
        for each hole and the lower face.

    Examples
    --------
    >>> fit_implant_to_mri_from_files("/path/to/file", "/path/to/directory")
    {'implant': <trimesh.Trimesh>, 1: <numpy.ndarray>, 2: <numpy.ndarray>}
    """
    if save_name is not None:
        save_path = Path(save_name)
        if save_path.is_dir():
            if mouse_name is not None:
                file_name = f"{mouse_name}_implant_fit.h5"
            else:
                file_name = "implant_fit.h5"
            save_file_path = save_path / file_name
        elif save_path.suffix == "h5":
            save_file_path = save_path
        else:
            raise ValueError("save_name must be a directory or an h5 file.")
        if not force and Path(save_file_path).is_file():
            raise FileExistsError(f"{save_file_path} already exists.")
    else:
        save_file_path = None

    implant_annotations = sitk.ReadImage(str(implant_annotations_file))
    hole_seg_dict = make_hole_seg_dict(implant_annotations)
    hole_mesh_dict = find_load_meshes(hole_directory, **find_load_mesh_kws)

    rotation_matrix, translation = fit_implant_to_mri(
        hole_seg_dict, hole_mesh_dict, **fit_kws
    )

    if save_file_path is not None:
        transform = rotation_matrix_to_sitk(
            rotation=rotation_matrix, translation=translation
        )
        if save_inverse:
            transform = transform.GetInverse()
        sitk.WriteTransform(transform, str(save_file_path))

    return rotation_matrix, translation



# %%
# Basic python functions that we need
from itertools import product
from pathlib import Path

import numpy as np
import pandas as pd
import SimpleITK as sitk
import trimesh

# Functions from aind_mri_utils
from aind_mri_utils import coordinate_systems as cs
from aind_mri_utils import rotations as rot
from aind_mri_utils.arc_angles import calculate_arc_angles
from aind_mri_utils.chemical_shift import (
    chemical_shift_transform,
    compute_chemical_shift,
)
from aind_mri_utils.file_io import simpleitk as mr_sitk
from aind_mri_utils.file_io import slicer_files as sf
from aind_mri_utils.file_io.obj_files import get_vertices_and_faces
from aind_mri_utils.meshes import load_newscale_trimesh
from aind_mri_utils.optimization import get_headframe_hole_lines
from matplotlib import cm
from scipy.spatial.transform import Rotation

# %%
# File Paths
mouse = "743700"
whoami = "galen"
if whoami == "galen":
    base_dir = Path("/mnt/aind1-vast/scratch/")
    base_save_dir = Path("/home/galen.lynch/")
elif whoami == "yoni":
    base_dir = Path("Y:/")
    base_save_dir = Path(
        "C:/Users/yoni.browning/OneDrive - Allen Institute/Desktop/"
    )
else:
    raise ValueError("Who are you again?")

headframe_model_dir = base_dir / "ephys/persist/data/MRI/HeadframeModels/"
probe_model_file = (
    headframe_model_dir / "dovetailtweezer_oneShank_centered_corrected.obj"
)  # "modified_probe_holder.obj"
annotations_path = base_dir / "ephys/persist/data/MRI/processed/{}/".format(
    mouse
)

headframe_path = headframe_model_dir / "TenRunHeadframe.obj"
holes_path = headframe_model_dir / "OneOff_HolesOnly.obj"


implant_holes_path = str(
    annotations_path / "{}_ImplantHoles.seg.nrrd".format(mouse)
)

image_path = str(
    annotations_path / "{}_100.nii.gz".format(mouse)
)  # '_100.nii.gz'))
labels_path = str(
    annotations_path / "{}_HeadframeHoles.seg.nrrd".format(mouse)
)  # 'Segmentation.seg.nrrd')#
brain_mask_path = str(
    annotations_path / ("{}_auto_skull_strip.nrrd".format(mouse))
)
manual_annotation_path = str(
    annotations_path / (f"{mouse}_ManualAnnotations.fcsv")
)
cone_path = (
    base_dir
    / "ephys/persist/Software/PinpointBuilds/WavefrontFiles/Cone_0160-200-53.obj"  # noqa E501
)

uw_yoni_annotation_path = (
    annotations_path / f"targets-{mouse}-transformed.fcsv"
)

newscale_file_name = headframe_path / "Centered_Newscale_2pt0.obj"
#


calibration_filename = "calibration_info_np2_2024_04_22T11_15_00.xlsx"
calibration_dir = (
    base_dir / "ephys/persist/data/probe_calibrations/CSVCalibrations/"
)
calibration_file = calibration_dir / calibration_filename
measured_hole_centers = annotations_path / "measured_hole_centers.xlsx"


# manual_hole_centers_file = annotations_path / 'hole_centers.mrk.json'

transformed_targets_save_path = annotations_path / (
    mouse + "TransformedTargets.csv"
)
test_probe_translation_save_path = str(
    base_save_dir / "test_probe_translation.h5"
)
transform_filename = str(annotations_path / (mouse + "_com_plane.h5"))
# Magic numbers
resolution = 100


# %%
# Handle inconsistant labeling
label_vol = sitk.ReadImage(labels_path)
odict = {k: label_vol.GetMetaData(k) for k in label_vol.GetMetaDataKeys()}
insert_underscores = (
    "_" in list(sf.find_seg_nrrd_header_segment_info(odict).keys())[0]
)

# Load the points on the headframe lines.
pts1, pts2, order = get_headframe_hole_lines(
    insert_underscores=insert_underscores, coordinate_system="LPS"
)

# order.remove('anterior_vertical')

image = sitk.ReadImage(image_path)


# Load the headframe
headframe, headframe_faces = get_vertices_and_faces(headframe_path)
headframe_lps = cs.convert_coordinate_system(
    headframe, "ASR", "LPS"
)  # Preserves shape!

# Load just the headframe holes
holes, holes_faces = get_vertices_and_faces(holes_path)
holes_faces = holes_faces[-1]
holes_lps = cs.convert_coordinate_system(holes, "ASR", "LPS")

# Load the brain mask
mask = sitk.ReadImage(brain_mask_path)
idxx = np.where(sitk.GetArrayViewFromImage(mask))
idx = np.vstack((idxx[2], idxx[1], idxx[0])).T
brain_pos = np.zeros(idx.shape)
brain_pos = np.vstack(
    [
        mask.TransformIndexToPhysicalPoint(idx[ii, :].tolist())
        for ii in range(idx.shape[0])
    ]
)
brain_pos = brain_pos[
    np.arange(0, brain_pos.shape[0], brain_pos.shape[0] // 1000)
]


# %%
# Load the computed transform
trans = mr_sitk.load_sitk_transform(
    transform_filename, homogeneous=True, invert=True
)[0]

# Get chemical shift from MRI image.
# Defaults are standard UW scans- set params for anything else.
chem_shift = compute_chemical_shift(image)
chem_shift_trans = chemical_shift_transform(chem_shift, readout="HF")
# -

# Read points
manual_annotation = sf.read_slicer_fcsv(manual_annotation_path)

# %%
# List targeted locations
preferred_pts = {
    # "LGN": manual_annotation["LGN"],
    "CCant": manual_annotation["CCant"],
    "CCpst": manual_annotation["CCpst"],
    "AntComMid": manual_annotation["AntComMid"],
    "GenFacCran": manual_annotation["GenFacCran2"],
}

hmg_pts = rot.prepare_data_for_homogeneous_transform(
    np.array(tuple(preferred_pts.values()))
)
chem_shift_annotation = hmg_pts @ trans.T @ chem_shift_trans.T
transformed_annotation = rot.extract_data_for_homogeneous_transform(
    chem_shift_annotation
)
target_nms = tuple(preferred_pts.keys())

for ii in range(transformed_annotation.shape[0]):
    print(
        "Target "
        + str(target_nms[ii])
        + " ML: "
        + str(-transformed_annotation[ii, 0])
        + " AP: "
        + str(-transformed_annotation[ii, 1])
        + " DV: "
        + str(transformed_annotation[ii, 2])
    )

ndf = pd.DataFrame(
    {
        "Target": target_nms,
        "ML": -transformed_annotation[:, 0],
        "AP": -transformed_annotation[:, 1],
        "DV": transformed_annotation[:, 2],
    }
)
ndf.to_csv(transformed_targets_save_path)

# %%
# If implant has holes that are segmented.
implant_vol = sitk.ReadImage(implant_holes_path)
odict = {k: implant_vol.GetMetaData(k) for k in implant_vol.GetMetaDataKeys()}
label_dict = sf.find_seg_nrrd_header_segment_info(odict)

implant_names = []
implant_targets = []
implant_pts = []

for ii, key in enumerate(label_dict.keys()):
    filt = sitk.EqualImageFilter()
    is_label = filt.Execute(implant_vol, label_dict[key])
    idxx = np.where(sitk.GetArrayViewFromImage(is_label))
    if len(idxx[0]) == 0:
        continue
    idx = np.vstack((idxx[2], idxx[1], idxx[0])).T
    implant_pos = np.vstack(
        [
            implant_vol.TransformIndexToPhysicalPoint(idx[ii, :].tolist())
            for ii in range(idx.shape[0])
        ]
    )
    implant_pts.append(implant_pos)
    implant_targets.append(np.mean(implant_pos, axis=0))
    this_key = key.split("-")[-1].split("_")[-1]
    implant_names.append(int(this_key))
implant_targets = np.vstack(implant_targets)


# %%
# Visualize Holes, list locations
transformed_brain = rot.extract_data_for_homogeneous_transform(
    np.dot(
        np.dot(
            rot.prepare_data_for_homogeneous_transform(brain_pos),
            chem_shift_trans,
        ),
        trans,
    )
)
transformed_implant = rot.extract_data_for_homogeneous_transform(
    np.dot(rot.prepare_data_for_homogeneous_transform(implant_targets), trans)
)


for ii in range(transformed_implant.shape[0]):
    print(
        "MRI "
        + str(implant_names[ii])
        + " ML: "
        + str(-transformed_implant[ii, 0])
        + " AP: "
        + str(-transformed_implant[ii, 1])
        + " DV: "
        + str(transformed_implant[ii, 2])
    )


# %%
preferred_pts = ["LGN", "CCant", "CCpst", "AntComMid", "GenFacCran"]
targets = list(preferred_pts.keys())
transformed_preferred = transformed_annotation

TARGET = []
HOLE = []
AP = []
AP_Range = []
RIG_AP = []
ML = []
ML_Range = []
TARGET_LOC = []
for tt in range(transformed_preferred.shape[0]):
    for hh in range(transformed_implant.shape[0]):
        insertion_vector = (
            transformed_implant[hh, :] - transformed_preferred[tt, :]
        )
        ap, ml = calculate_arc_angles(
            transformed_preferred[tt, :],
            transformed_implant[hh, :],
            ap_offset=0,
        )

        radius = 0.3
        theta = np.deg2rad(np.arange(0, 360, 1))
        a = transformed_implant[hh, 0] + radius * np.cos(theta)
        b = transformed_implant[hh, 1] + radius * np.sin(theta)
        circle = np.vstack(
            [a, b, np.ones(b.shape) * transformed_implant[hh, 2]]
        ).T

        this_ML = []
        this_AP = []

        for jj in range(len(a)):

            this_ap, this_ml = calculate_arc_angles(
                transformed_preferred[tt, :], circle[jj, :], ap_offset=0
            )
            this_ML.append(this_ml)
            this_AP.append(this_ap)

        ML_Range.append(np.abs(np.max(this_ML) - np.min(this_ML)) / 2)
        AP_Range.append(np.abs(np.max(this_AP) - np.min(this_AP)) / 2)

        if False:
            continue
        else:
            TARGET.append(targets[tt])
            HOLE.append(implant_names[hh])
            ML.append(ml)
            AP.append(ap)
            RIG_AP.append(ap + 14)
            TARGET_LOC.append(transformed_preferred[tt, :])


df = pd.DataFrame(
    {
        "target": TARGET,
        "hole": HOLE,
        "rig_ap": RIG_AP,
        "ml": ML,
        "ap": AP,
        "ML_range": ML_Range,
        "AP_range": AP_Range,
        "target_loc": TARGET_LOC,
    }
)


# %%
ap_wiggle = 1
ap_min = 16
ml_min = 16
rotation_inc = 45
valid = np.zeros([df.shape[0], df.shape[0]], dtype=bool)
# mesh = load_newscale_model(move_down=.5,filename=probe_model_file)
mesh = load_newscale_trimesh(probe_model_file, move_down=0.5)

cone = trimesh.load_mesh(cone_path)
cone.vertices = cs.convert_coordinate_system(cone.vertices, "ASR", "LPS")

hf = trimesh.Trimesh()
hf.vertices = headframe_lps
hf.faces = headframe_faces
CM = trimesh.collision.CollisionManager()
CM.add_object("cone", cone)
# check the validity of every insertions. This is order N**2, so may need some
# thought.

for ii, row1 in df.iterrows():
    for jj, row2 in df.iterrows():
        if ii == jj or ii < jj:
            continue
        # elif row1.target == row2.target:
        #    continue
        elif row1.hole == row2.hole:
            continue
        elif (np.abs(row1.ap - row2.ap) < ap_wiggle) and (
            np.abs(row1.ml - row2.ml) < ml_min
        ):
            continue
        elif (np.abs(row1.ap - row2.ap) < ap_min) and (
            np.abs(row1.ap - row2.ap) > ap_wiggle
        ):

            continue
        else:
            valid[ii, jj] = True
valid = np.bitwise_or(valid, valid.T)

# -

df[(df.target == "GenFacCran")]

# %%
match_insertions = [68, 59, 67]
works_for_all = set(np.where(valid[match_insertions[0], :])[0])

for ii in range(0, len(match_insertions)):
    works_for_all = (
        works_for_all
        & set(np.where(valid[match_insertions[ii], :])[0])
        & set(np.where(df.target == "GenFacCran")[0])
    )

df.iloc[np.concatenate([match_insertions, list(works_for_all)])]


# %%
def transform_matrix_from_angles_and_target(AP, ML, Target, degrees=True):
    R = Rotation.from_euler(
        "XYZ", np.array([np.deg2rad(AP), np.deg2rad(ML), 0])
    ).as_matrix()
    T = np.zeros([4, 4])
    T[:3, :3] = R
    T[0:3, 3] = Target
    return T


insertion_list = match_insertions


new_mesh = trimesh.Trimesh()
new_mesh.vertices = headframe_lps
new_mesh.faces = headframe_faces


cone = trimesh.load_mesh(cone_path)
cone.vertices = cs.convert_coordinate_system(cone.vertices, "ASR", "LPS")


S = trimesh.scene.Scene([new_mesh])
S.add_geometry(cone)

meshes = [
    trimesh.creation.uv_sphere(radius=0.25)
    for i in range(len(transformed_implant))
]
for i, m in enumerate(meshes):
    m.apply_translation(transformed_implant[i, :])
    m.visual.vertex_colors = [255, 0, 255, 255]
    S.add_geometry(m)

meshes = [
    trimesh.creation.uv_sphere(radius=0.25)
    for i in range(len(transformed_preferred))
]
for i, m in enumerate(meshes):
    m.apply_translation(transformed_preferred[i, :])
    m.visual.vertex_colors = trimesh.visual.random_color()
    S.add_geometry(m)


S.add_geometry(new_mesh)

for ii in match_insertions:
    T1 = transform_matrix_from_angles_and_target(
        df.ap[ii], -df.ml[ii], df.target_loc[ii]
    )
    S.add_geometry(rot.apply_transform_to_trimesh(mesh.copy(), T1))

S.set_camera([0, 0, 0], distance=150, center=[0, 0, 0])
S.show()
# -

#


# %%

# Tools for collsion testings
mesh = load_newscale_trimesh(
    probe_model_file,
    -0.2,
)

angle_ranges = [np.arange(0, 360, 45)] * len(match_insertions)

angle_sets = [x for x in product(*angle_ranges)]

# Load the stuff that doesn't change on each iteration
cone = trimesh.load_mesh(cone_path)
cone.vertices = cs.convert_coordinate_system(cone.vertices, "ASR", "LPS")
CM = trimesh.collision.CollisionManager()
new_mesh = trimesh.Trimesh()
new_mesh.vertices = headframe_lps
new_mesh.faces = headframe_faces
# CM.add_object("cone", cone)

insert_list = match_insertions

for this_angle in angle_sets:

    for this_insertion in range(len(match_insertions)):
        # Mesh1
        this_mesh = mesh.copy()
        TA = trimesh.transformations.euler_matrix(
            0, 0, np.deg2rad(this_angle[this_insertion])
        )
        TB = transform_matrix_from_angles_and_target(
            df.ap[insert_list[this_insertion]],
            -df.ml[insert_list[this_insertion]],
            df.target_loc[insert_list[this_insertion]],
        )  # my ml convention is backwards

        rot.apply_transform_to_trimesh(this_mesh, TA)
        rot.apply_transform_to_trimesh(this_mesh, TB)
        CM.add_object(f"mesh{this_insertion}", this_mesh)

    if not CM.in_collision_internal(False, False):
        print("pass :" + str(this_angle))
        break
    else:
        print("fail :" + str(this_angle))

    for this_insertion in range(len(match_insertions)):
        CM.remove_object(f"mesh{this_insertion}")


S = trimesh.scene.Scene([new_mesh])

S.add_geometry(new_mesh)


cstep = (256) // (len(match_insertions))

for this_insertion in range(len(match_insertions)):
    this_mesh = mesh.copy()
    TA = trimesh.transformations.euler_matrix(
        0, 0, np.deg2rad(this_angle[this_insertion])
    )
    TB = transform_matrix_from_angles_and_target(
        df.ap[insert_list[this_insertion]],
        -df.ml[insert_list[this_insertion]],
        df.target_loc[insert_list[this_insertion]],
    )  # my ml convention is backwards

    rot.apply_transform_to_trimesh(this_mesh, TA)
    rot.apply_transform_to_trimesh(this_mesh, TB)
    this_mesh.visual.vertex_colors = (
        np.array(cm(this_insertion * cstep)) * 255
    ).astype(int)
    S.add_geometry(this_mesh)


meshes = [
    trimesh.creation.uv_sphere(radius=0.25)
    for i in range(len(transformed_implant))
]
for i, m in enumerate(meshes):
    m.apply_translation(transformed_implant[i, :])
    m.visual.vertex_colors = [255, 0, 255, 255]
    S.add_geometry(m)

meshes = [
    trimesh.creation.uv_sphere(radius=0.25)
    for i in range(len(transformed_preferred))
]
for i, m in enumerate(meshes):
    m.apply_translation(transformed_preferred[i, :])
    m.visual.vertex_colors = trimesh.visual.random_color()
    S.add_geometry(m)
S.add_geometry(cone)

S.show()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.3
#   kernelspec:
#     display_name: Python 3.9.12 ('base')
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Probe calibration and target transformation notebook
#
# This notebook is used to calibrate probes and transform targets from
# bregma-relative mm to manipulator coordinates
#
# # How to use this notebook
# 1. Set the mouse ID in the cell two below.
# 2. Set the path to the calibration file with the probe data.
# 3. Set the path to the target file.
# 4. Optionally set `fit_scale` to `True` if you want to fit the scale
# parameters as well. This is not recommended unless you have a good reason to
# do so. It does not guarantee that the error will be lower.
# 5. Run the next three cells to get the transformed targets, and see which
# targets are available
# 6. Configure the experiment by assigning each probe that you want to use to a
# target in the target file and specify the overshoot in µm. If you have
# targets that are not in the target file, you can specify them manually.
# 7. Run the next cell to fit the rotation parameters. If `verbose` is set to
# `True`, the mean and maximum error for each probe will be printed, as well as
# the predicted probe coordinates for each reticle coordinate with error for
# that coordinate.
# 8. Run the last cell to get the transformed targets in manipulator
# coordinates

# %%
from pathlib import Path

import numpy as np
import pandas as pd

# %matplotlib inline
from aind_mri_utils import reticle_calibrations as rc

# %%
# Set file paths and mouse ID here

# Calibration File with probe data
mouse_id = "727354"
basepath = Path("/mnt/aind1-vast/scratch/")
calibration_dir = (
    basepath / "ephys/persist/data/probe_calibrations/CSVCalibrations/"
)

# Target file with transformed targets
target_dir = basepath / f"ephys/persist/data/MRI/processed/{mouse_id}/"
cal_file = calibration_dir / "calibration_info_np2_2024_08_05T14_34_00.xlsx"
target_file = target_dir / f"{mouse_id}_TransformedTargets.csv"

# Whether to fit the scale parameters as well. This is not recommended unless
# you have a good reason to do so.  Does not guarantee that the error will be
# lower
fit_scale = False

# Whether to print the mean and maximum error for each probe and the predicted
# probe coordinates for each reticle coordinate with error for that coordinate
verbose = True


# %%
def _round_targets(target, probe_target):
    target_rnd = np.round(target, decimals=2)
    probe_target_and_overshoot_rnd = (
        np.round(2000 * probe_target_and_overshoot) / 2
    )
    return target_rnd, probe_target_and_overshoot_rnd


# %%
target_df = pd.read_csv(target_file)
target_df = target_df.set_index("point")
# %% [markdown]
# ## Transformed targets
# print the transformed targets to see which targets are available
# %%
print(target_df)

# %% [markdown]
# ## Configure experiment
# Assign each probe that you want to use to a target in the target file and
# specify the overshoot in µm. The format should be
# ```python
# targets_and_overshoots_by_probe = {
#     probe_id: (target_name, overshoot), # overshoot in µm
#     ...
# }
# ```
# Where each `probe_id` is the ID of a probe in the calibration file,
# `target_name` is the name of the target in the target file, and `overshoot`
# is the overshoot in µm.
#
# If you have targets that are not in the target file, you can specify them
# manually. The format should be
#
# ```python manual_bregma_targets_by_probe = { probe_id: [x, y, z], ...  } ```
# where `[x, y, z]` are the coordinates in mm.  %% Set experiment configuration
# here

# Names of targets in the target file and overshoots
# targets_and_overshoots_by_probe = {probe_id: (target_name, overshoot), ...}
# overshoot in µm
targets_and_overshoots_by_probe = {
    46110: ("AntComMid", 500),
    46100: ("GenFacCran2", 500),
}
# Targets in bregma-relative coordinates not in the target file
# manual_bregma_targets_by_probe = {probe_id: [x, y, z], ...}
# x y z in mm
manual_bregma_targets_by_probe = {
    # 46110: [0, 0, 0], # in mm!
}


# %%
(
    adjusted_pairs_by_probe,
    global_offset,
    global_rotation_degrees,
    reticle_name,
) = rc.read_reticle_calibration(cal_file)

# %% [markdown]
# ## Fit rotation parameters
# Fit the rotation parameters and optionally the scale parameters. If `verbose`
# is set to `True`, the mean and maximum error for each probe will be printed,
# as well as the predicted probe coordinates for each reticle coordinate with
# error for that coordinate.
#
# Note: the reticle coordinates are in mm, as are the probe coordinates. The
# errors are in µm.
#
# The reticle coordinate displayed will NOT have the global offset applied.
# However, the scaling factor will have been applied.
# %%
# Calculate the rotation parameters and display errors if verbose is set to
# True

rotations = dict()
translations = dict()
if fit_scale:
    scale_vecs = dict()
    for probe, (reticle_pts, probe_pts) in adjusted_pairs_by_probe.items():
        (rotation, scale, translation, _) = rc.fit_rotation_params(
            reticle_pts, probe_pts, find_scaling=True
        )
        rotations[probe] = rotation
        scale_vecs[probe] = scale
        translations[probe] = translation
else:
    for probe, (reticle_pts, probe_pts) in adjusted_pairs_by_probe.items():
        rotation, translation, _ = rc.fit_rotation_params(
            reticle_pts, probe_pts, find_scaling=False
        )
        rotations[probe] = rotation
        translations[probe] = translation

for probe, (reticle_pts, probe_pts) in adjusted_pairs_by_probe.items():
    if fit_scale:
        scale = scale_vecs[probe]
    else:
        scale = None
    predicted_probe_pts = rc.transform_reticle_to_probe(
        reticle_pts, rotations[probe], translations[probe], scale
    )
    # in µm
    errs = 1000 * np.linalg.norm(predicted_probe_pts - probe_pts, axis=1)
    if verbose:
        print(
            f"Probe {probe}: Mean error {errs.mean():.2f} µm, "
            f"max error {errs.max():.2f} µm"
        )
        original_reticle_pts = reticle_pts - global_offset
        for i in range(len(errs)):
            rounded_pred = np.round(predicted_probe_pts[i], decimals=2)
            print(
                f"\tReticle {original_reticle_pts[i]} -> "
                f"Probe {probe_pts[i]}: predicted {rounded_pred} "
                f"error {errs[i]:.2f} µm"
            )

# %% [markdown]
# ## Probe targets in manipulator coordinates
# Get the transformed targets in manipulator coordinates using the fitted
# calibration parameters and the experiment configuration set in the previous
# cells.


# %%
# Print the transformed targets in manipulator coordinates

for probe, (target_name, overshoot) in targets_and_overshoots_by_probe.items():
    if probe not in rotations:
        print(f"Probe {probe} not in calibration file")
        continue
    target = target_df.loc[target_name].to_numpy()
    overshoot_arr = np.array([0, 0, overshoot / 1000])
    if fit_scale:
        scale = scale_vecs[probe]
    else:
        scale = None
    probe_target = rc.transform_reticle_to_probe(
        target, rotations[probe], translations[probe], scale
    )
    probe_target_and_overshoot = probe_target + overshoot_arr
    target_rnd, probe_target_and_overshoot_rnd = _round_targets(
        target, probe_target_and_overshoot
    )
    print(
        f"Probe {probe}: Target {target_name} {target_rnd} (mm) "
        f"-> manipulator coord. {probe_target_and_overshoot_rnd} (µm) "
        f"w/ {overshoot} µm overshoot"
    )
for probe, target in manual_bregma_targets_by_probe.items():
    if probe not in rotations:
        print(f"Probe {probe} not in calibration file")
        continue
    target_arr = np.array(target)
    if fit_scale:
        scale = scale_vecs[probe]
    else:
        scale = None
    probe_target = rc.transform_reticle_to_probe(
        target_arr, rotations[probe], translations[probe], scale
    )
    target_rnd, probe_target_rnd = _round_targets(target_arr, probe_target)
    print(
        f"Probe {probe}: Manual target {target_rnd} (mm) -> manipulator "
        f"coord. {probe_target_rnd} (µm)"
    )
